package main

import (
	"fmt"
	"regexp"
	"sort"
	"strconv"
	"strings"

	"github.com/go-git/go-git/v5/plumbing"
	"github.com/go-git/go-git/v5/plumbing/object"
	"github.com/go-git/go-git/v5/plumbing/storer"

	"github.com/go-git/go-git/v5"
)

var versionRegex = regexp.MustCompile(`^v?(\d+)\.(\d+)\.(\d+).*$`)

type SemanticVersion struct {
	Major int
	Minor int
	Patch int
	Tag   string
}

func (v SemanticVersion) String() string {
	return fmt.Sprintf("%d.%d.%d", v.Major, v.Minor, v.Patch)
}

func ParseSemanticVersion(tag string) SemanticVersion {
	matches := versionRegex.FindStringSubmatch(tag)
	major, _ := strconv.Atoi(matches[1])
	minor, _ := strconv.Atoi(matches[2])
	patch, _ := strconv.Atoi(matches[3])
	return SemanticVersion{major, minor, patch, tag}
}

func (v SemanticVersion) IsLessThan(w SemanticVersion) bool {
	if v.Major != w.Major {
		return v.Major < w.Major
	}
	if v.Minor != w.Minor {
		return v.Minor < w.Minor
	}
	if v.Patch != w.Patch {
		return v.Patch < w.Patch
	}
	return false
}

type CommitType int

const (
	OtherCommit CommitType = iota
	FeatCommit
	BreakingCommit
	FixCommit
)

func (c CommitType) String() string {
	switch c {
	case FeatCommit:
		return "feat"
	case BreakingCommit:
		return "breaking"
	case FixCommit:
		return "fix"
	default:
		return "other"
	}
}

func GetCommitType() (CommitType, error) {

	latestTag, err := getLatestSemanticTag()
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("latest found tag:", latestTag)
	}

	tagCommit, err := getTagCommit(latestTag.Tag)
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("tag commit:", tagCommit)
	}

	commitType, err := findNextTagType(gitRepo, tagCommit, verbosity)
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("commit type:", commitType)
	}

	return commitType, nil
}

func getLatestSemanticTag() (SemanticVersion, error) {
	tagList, err := getSemanticTags(gitRepo)
	if err != nil {
		return SemanticVersion{}, err
	}

	if len(tagList) == 0 {
		return SemanticVersion{}, fmt.Errorf("no semantic tags found")
	}

	return tagList[len(tagList)-1], nil
}

func getTagCommit(tag string) (*object.Commit, error) {

	var r = gitRepo

	tags, err := r.Tags()
	if err != nil {
		return nil, fmt.Errorf("failed to get tags: %v", err)
	}

	var commitHash *object.Commit
	err = tags.ForEach(func(t *plumbing.Reference) error {
		if t.Name().Short() != tag {
			return nil
		}

		// Resolve the tag to a commit
		obj, err := r.TagObject(t.Hash())
		var hash plumbing.Hash
		if err == nil {
			// This is an annotated tag
			hash = obj.Target
		} else {
			// This is a lightweight tag or an error occurred
			hash = t.Hash()
		}

		commitHash, err = r.CommitObject(hash)

		return ErrStopIteration

	})

	if err != nil && err != ErrStopIteration {
		return nil, fmt.Errorf("failed to iterate over tags: %v", err)
	}

	if commitHash == nil {
		return nil, fmt.Errorf("tag '%s' not found in commit log", tag)
	}

	return commitHash, nil
}

func getSemanticTags(r *git.Repository) ([]SemanticVersion, error) {
	tags, err := r.Tags()
	if err != nil {
		return nil, fmt.Errorf("failed to get tags: %v", err)
	}

	headRef, err := r.Head()
	if err != nil {
		return nil, fmt.Errorf("failed to get branch: %v", err)
	}

	commitIter, err := r.Log(&git.LogOptions{From: headRef.Hash()})
	if err != nil {
		return nil, fmt.Errorf("failed to get commit history: %v", err)
	}

	commitHistory := make(map[plumbing.Hash]bool)
	err = commitIter.ForEach(func(c *object.Commit) error {
		commitHistory[c.Hash] = true
		return nil
	})

	if err != nil {
		return nil, fmt.Errorf("failed to iterate over commits: %v", err)
	}

	var tagList []SemanticVersion
	err = tags.ForEach(func(tag *plumbing.Reference) error {
		tagName := tag.Name().Short()

		if verbosity {
			fmt.Println("found tag: ", tagName)
		}

		// Resolve the tag to a commit
		var commitHash plumbing.Hash
		obj, err := r.TagObject(tag.Hash())
		if err == nil {
			// This is an annotated tag
			commitHash = obj.Target
		} else {
			// This is a lightweight tag or an error occurred
			commitHash = tag.Hash()

			if verbosity {
				fmt.Println("tag " + tagName + " is a lightweight tag, you should use an annotated tag")
			}

		}

		if _, exists := commitHistory[commitHash]; !exists {
			return nil
		}

		if versionRegex.MatchString(tagName) {
			tagList = append(tagList, ParseSemanticVersion(tagName))
		}

		return nil
	})

	if err != nil {
		return nil, fmt.Errorf("failed to iterate over tags: %v", err)
	}

	sort.Slice(tagList, func(i, j int) bool {
		return tagList[i].IsLessThan(tagList[j])
	})

	if verbosity {
		fmt.Println("tag list: ", tagList)
	}

	return tagList, nil
}

func findNextTagType(r *git.Repository, tagCommit *object.Commit, verbosity bool) (CommitType, error) {

	cIter, err := r.Log(&git.LogOptions{})

	if err != nil {
		return OtherCommit, fmt.Errorf("failed to get commit log from tag: %v", err)
	}

	var commitType CommitType = OtherCommit
	counter := 0

	err = cIter.ForEach(func(commit *object.Commit) error {
		counter++

		if commit.Hash == tagCommit.Hash {
			return storer.ErrStop
		}

		message := strings.TrimSpace(commit.Message)
		if strings.HasPrefix(message, "feat") {
			commitType = FeatCommit
			if verbosity {
				fmt.Printf("Found 'feat' commit after %d commits\n", counter)
			}
			return storer.ErrStop
		} else if strings.HasPrefix(message, "fix") {
			if commitType < FixCommit {
				commitType = FixCommit
				if verbosity {
					fmt.Printf("Found 'fix' commit after %d commits\n", counter)
				}
			}
		} else if containsBreakingChangeFooter(message) {
			commitType = BreakingCommit
			if verbosity {
				fmt.Printf("Found breaking change after %d commits\n", counter)
			}
			return storer.ErrStop
		}
		return nil
	})

	if err != nil {
		return OtherCommit, fmt.Errorf("failed to iterate over commit log: %v", err)
	}

	return commitType, nil
}

func containsBreakingChangeFooter(message string) bool {
	lines := strings.Split(message, "\n")
	for i := len(lines) - 1; i >= 0; i-- {
		if strings.TrimSpace(lines[i]) == "" {
			break
		}
		if strings.Contains(strings.ToLower(lines[i]), "breaking change:") {
			return true
		}
	}
	return false
}
